{$ifdef cppbuilder}
{$ObjExportAll On}
{$endif}
{
  Unit type (Component, Formless unit, Ancestor form, Application form,...)
  ------------------------------------------------------------------------------
  Component

  Purpose
  ------------------------------------------------------------------------------
  Use the persistor to save and load the values of specified properties to
  a file or the registry.

  History
  ------------------------------------------------------------------------------
  1998-05-00 THk Created


  Copyright
  ------------------------------------------------------------------------------
  (C) 1998 - Thor Kristmundsson

}
unit Persistor;
interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls,
  Forms, Dialogs, typinfo, dsgnintf;

type
  TJOptions=(joAutoStore,joAutoLoad,joStoreEmpty,joLoadEmpty);
  TJOptionsSet=set of TJOptions;
  TPersistorStorage=(jsInternal,jsRegistry,jsFile);

type
  TPersistor = class(TComponent)
  private
    { Private declarations }
    ObjectInstance : Pointer;
    oldOwnerWndPrc : Pointer;
    procedure newOwnerWndPrc(var Message: TMessage);
  protected
    { Protected declarations }
    fExcludeMask:integer;
    fVisited:TList;
    fOptions:TJOptionsSet;
  public
    { Public declarations }
    fConfig:String;
    fStorage:TPersistorStorage;
    fPath:string;
    fWhichProperties:TStrings;
    fVault:TStrings;
    fPropValues:TStringList;
    fCustomProps:Boolean;
    constructor create(AOwner:TComponent); override;
    destructor destroy; override;
    function substfilename:string; virtual;
    procedure SetWhichProperties(Value:TStrings);
    procedure SetVault(Value:TStrings);
    function CollectWhichProperties(cls:PTypeInfo):String; virtual;
    function GetPropValAsString(p:PPropInfo; obj:TObject):String; virtual;
    procedure SetPropValAsString(p:PPropInfo; obj:TObject; val:string); virtual;
    procedure traverse(prefix:string;Obj:TObject;isWriting:boolean); virtual;
    procedure traverseAll(prefix:string;Obj:TObject;isWriting:boolean); virtual;
    procedure SaveToFile; virtual;
    procedure LoadFromFile; virtual;
    procedure FromStorage; virtual;
    procedure ToStorage; virtual;
    procedure FromForm; virtual;
    procedure ToForm; virtual;
    procedure RunDialog(ACaption:String;ShowWhichProperties:Boolean); virtual;
    procedure RunDialogAttached(AParent:TWinControl;AAlign:TAlign; ACaption:String;ShowWhichProperties:Boolean);
    procedure Loaded; override;
  published
    { Published declarations }
    property Storage:TPersistorStorage read fStorage write fStorage;
    property ExcludeMask:integer read fExcludeMask write fExcludeMask;
    property WhichProperties:TStrings read fWhichProperties write SetWhichProperties;
    property Vault:TStrings read fVault write SetVault;
    property Path:string read fPath write fPath;
    property Config:string read fConfig write fConfig;
    property Options:TJOptionsSet read fOptions write fOptions;
  end;

  { Component editor to edit the project file. }
  TPersistorEditor = class(TComponentEditor)
  public
    function GetVerb(Index: Integer): string; override;
    function GetVerbCount: Integer; override;
    procedure ExecuteVerb(Index: Integer); override;
    procedure Edit; override;
  end;

procedure Register;

implementation
{$R Persistor.dcr}
uses registry, frmuPersistor, utlStrings;

function TPersistorEditor.GetVerbCount: Integer;
begin
  Result:=3;
end;

function TPersistorEditor.GetVerb(Index: Integer): string;
begin
  case index of
  0: Result := 'Load values';
  1: Result := 'Save values';
  2: Result := 'Edit';
  end;
end;

procedure TPersistorEditor.ExecuteVerb(Index: Integer);
begin
  with component as TPersistor do
    case index of
    0: LoadFromFile;
    1: SaveToFile;
    2: RunDialog(Name,true);
    end;
end;

procedure TPersistorEditor.Edit;
begin
  with component as TPersistor do
    RunDialog(name,true);
end;


function tryFrontToken(var s:string; var token:string):boolean;
const lim=[',']; whsp=[' '];
begin
  token:='';
  result:=false;
  while (s<>'') and (s[1] in whsp) do
    delete(s,1,1);
  while (s<>'') and not (s[1] in lim) do
  begin
    result:=true;
    token:=token+s[1];
    delete(s,1,1);
  end;
  if (s<>'') and (s[1] in lim) then
    delete(s,1,1);
end;

procedure TPersistor.RunDialog(ACaption:String;ShowWhichProperties:Boolean);
begin
  RunDialogAttached(nil,alNone,ACaption,ShowWhichProperties);
end;

procedure TPersistor.RunDialogAttached(AParent:TWinControl;AAlign:TAlign; ACaption:String;ShowWhichProperties:Boolean);
var frm:TFrmPersistor;
begin
  frm:=TFrmPersistor.Create(Application);
  frm.Persistor:=Self;
  frm.memWhichProperties.Text:=WhichProperties.Text;
  if ACaption<>'' then
    frm.Caption:=ACaption;
  if not ShowWhichProperties then
  begin
    frm.memWhichProperties.Visible:=False;
    frm.StaticText2.Visible:=False;
    frm.Splitter1.Visible:=False;
    frm.btnCancel.Visible:=False;
  end;
  if AParent=nil then
  begin
    frm.ShowModal;
    frm.Free;
  end
  else
  begin
    frm.Parent:=AParent;
    frm.BorderStyle:=bsNone;
    frm.Align:=AAlign;
    frm.Show;
  end;
end;

constructor TPersistor.create(AOwner:TComponent);
begin
  inherited Create(AOwner);
  fWhichProperties:=TStringList.Create;
  fVault:=TStringList.Create;
  fPropValues:=TStringList.Create;
  fVisited:=TList.Create;
  if not (csDesigning in ComponentState) then
  begin
    ObjectInstance := MakeObjectInstance(newOwnerWndPrc);
    oldOwnerWndPrc := Pointer(GetWindowLong((Owner as TForm).Handle, GWL_WNDPROC));
    SetWindowLong((Owner as TForm).Handle, GWL_WNDPROC, LongInt(ObjectInstance));
  end; {}
end;

procedure TPersistor.Loaded;
begin
  if not (csDesigning in ComponentState) then
    if joAutoLoad in fOptions then
      LoadFromFile;
end;

destructor TPersistor.Destroy;
begin
  fWhichProperties.Free;
  fVault.Free;
  fPropValues.Free;
  fVisited.Free;
  inherited;
end;

procedure TPersistor.newOwnerWndPrc(var Message: TMessage);
begin
  with TMessage(Message) do
  begin
    if Message.Msg=WM_CLOSE then
    begin
      if not (csDesigning in ComponentState) then
        if joAutoStore in fOptions then
          SaveToFile;
      if (ObjectInstance <> Nil) then
        FreeObjectInstance(ObjectInstance);
      SetWindowLong((Owner as TForm).Handle, GWL_WNDPROC, LongInt(oldOwnerWndPrc));
    end;
    Result := CallWindowProc(oldOwnerWndPrc, (Owner as TForm).Handle, Msg, WParam, LParam);
  end;
end;

function TPersistor.substfilename:string;
var t:PTypeInfo;
begin
  t:=Owner.ClassInfo;
  result:=rxRpl(fPath,'<Config>',Config,false);
  result:=rxRpl(result,'<Formtype>',t^.name,false);
  result:=rxRpl(result,'<Formname>',
    getPropValAsString(GetPropInfo(Owner.Classinfo,'Name'),owner),true);
  result:=rxRpl(result,'<Caption>',
    getPropValAsString(GetPropInfo(Owner.Classinfo,'Caption'),owner),true);
end;

procedure TPersistor.SetWhichProperties(Value:TStrings);
begin
  WhichProperties.Assign(Value);
end;

procedure TPersistor.SetVault(Value:TStrings);
begin
  Vault.Assign(Value);
end;

function RemoveEl(list,el:string):string;
begin
  result:=rxRpls(list, [el,',,'], ['',','],true);
end;

{
  Find the properties to save for an object or its ancestors.
  A minus preceding a property name negate an ancestral property of
  that name. E.g.:
    TComponent=Left,Top,Width,Height
    TButton=Caption,Hint,-Width
  will result in the list Left,Top,Height,Caption,Hint for TButton
}
function TPersistor.CollectWhichProperties(cls:PTypeInfo):String;
var aname,props,prop:string;
begin
  aname:=cls.name;
  if aname<>'TObject' then
    result:=CollectWhichProperties(GetTypeData(cls).ParentInfo^);
  props:=fWhichProperties.Values[aname];
  if props<>'' then
    while TryFrontToken(props,prop) do
      if prop<>'' then
        if prop[1]='-' then
          result:=removeEl(result,copy(prop,2,100))
        else
          result:=result+','+prop;
  if (result<>'') and (result[1]=',') then
    delete(result,1,1);
end;

// quote control characters such as CR and LF
function quote(s:string):string;
var i:integer;
begin
  result:='';
  for i:=1 to length(s) do
    if s[i]<' ' then
      result:=result+'^'+char(ord('A')+ord(s[i]))
    else
      result:=result+s[i];
end;

// unquote control caracters
function unquote(s:string):string;
var i,len:integer;
begin
  result:='';
  len:=length(s);
  i:=1;
  while i<len do
  begin
    if (s[i]='^') and (s[i+1] in ['A'..'Z']) then
    begin
      result:=result+char(ord(s[i+1])-ord('A'));
      inc(i);
    end
    else
      result:=result+s[i];
    inc(i);
  end;
  if i=len then
    result:=result+s[len];
end;

function TPersistor.GetPropValAsString(p:PPropInfo; obj:TObject):String;
begin
  case p.proptype^.kind of
  tkString,tkLString: result:=getStrProp(obj,p);
  tkInteger,tkEnumeration,tkSet: result:=IntToStr(getOrdProp(obj,p));
  tkFloat: result:=floatToStr(getFloatProp(obj,p));
  end;
end;

function TryStrToInt(s:string;var i:integer):boolean;
var code:integer;
begin
  val(s,i,code);
  result:=code=0;
end;

function TryStrToFloat(s:string;var x:double):boolean;
var code:integer;
begin
  val(s,x,code);
  result:=code=0;
end;

procedure TPersistor.SetPropValAsString(p:PPropInfo; obj:TObject; val:string);
var n:integer; x:double;
begin
  try
    case p.proptype^.kind of
    tkString,tkLString: SetStrProp(obj,p,Val);
    tkInteger,tkEnumeration,tkSet:
      if tryStrToInt(val,n) then
        SetOrdProp(obj,p,n);
    tkFloat:
      if tryStrToFloat(val,x) then
        SetFloatProp(obj,p,x);
    end;
  except
  end;
end;

procedure TPersistor.traverse(prefix:string;Obj:TObject;isWriting:boolean);
var props,prop,key,value:string; i:integer; p:PPropInfo;
begin
  if Obj=nil then
    exit;
  if fVisited.IndexOf(Obj)>=0 then
    exit;
  fVisited.Add(Obj);

  p:=GetPropInfo(obj.ClassInfo,'Tag');
  if p<>nil then
    if (GetOrdProp(obj,p) and fExcludeMask)<>0 then
      exit;

  p:=GetPropInfo(obj.ClassInfo,'Name');
  if p<>nil then
    prefix:=prefix+'.'+GetPropValAsString(p,obj);

  props:=CollectWhichProperties(obj.ClassInfo);

  while tryFrontToken(Props,prop) do
  begin
    key:=prefix+'.'+prop;
    p:=GetPropInfo(obj.Classinfo,prop);
    if p<>nil then
      if (p.proptype^.kind=tkClass) then
        traverse(prefix+'.'+p.name,TObject(GetOrdProp(Obj,p)),isWriting)
      else if isWriting then
      begin
        value:=GetPropValAsString(p,obj);
        if (value<>'') or (joStoreEmpty in Options) then
          fPropValues.Values[key]:=quote(value)
      end
      else
      begin
        value:=unquote(fPropValues.Values[key]);
        if (value<>'') or (joLoadEmpty in Options) then
          SetPropValAsString(p,obj,value);
      end;
  end;

  // Traverse various structures here

  if obj is TForm then with obj as TForm do
    for i:=0 to ComponentCount-1 do
      traverse(prefix,components[i],isWriting);

  if obj is TCollection then with obj as TCollection do
    for i:=0 to Count-1 do
      traverse(format('%s[%d]',[prefix,i]),Items[i],isWriting);

  if obj is TStrings then with obj as TStrings do
    if iswriting then
    begin
      value:=quote(Text);
      if (value<>'') or (joStoreEmpty in Options) then
        fPropValues.add(format('%s.Text=%s',[prefix,quote(Text)]))
    end
    else
    begin
      value:=unquote(fPropValues.Values[format('%s.Text',[prefix])]);
      if (value<>'') or (joLoadEmpty in Options) then
        Text:=value;
    end;

  (*
  begin
    if Count=0 then // Correct the Count (vcl bug).
    begin
      s:=Text;
      Clear;
      Text:=s;
    end;
    for i:=0 to Count-1 do
      if isWriting then
        fPropValues.add(format('%s[%d]=%s',[prefix,i,Strings[i]]))
      else
        Strings[i]:=fPropValues.Values[format('%s[%d]',[prefix,i])];
  end;
  *)
end;

// Traverse all subobject properties of an object
procedure TPersistor.TraverseAll(Prefix:string; Obj: TObject; isWriting:boolean);
var PropInfos:PPropList; p:PPropInfo; n, i: Integer;
begin
  n := GetPropList(Obj.ClassInfo, tkAny, nil);
  GetMem(PropInfos, n*SizeOf(PPropInfo));
  try
    GetPropList(Obj.ClassInfo, tkAny, PropInfos);
    for i := 0 to n-1 do
    begin
      p:=PropInfos[i];
      if p.proptype^.kind=tkClass then
        traverse(prefix+'.'+p.name,TObject(GetOrdProp(Obj,p)),isWriting)
    end;
  finally
    FreeMem(PropInfos, n*SizeOf(PPropInfo));
  end;
end;


procedure TPersistor.SaveToFile;
begin
  FromForm;
  ToStorage;
end;

procedure TPersistor.LoadFromFile;
begin
  FromStorage;
  ToForm;
end;

procedure TPersistor.FromForm;
begin
  fVisited.Clear;
  fPropValues.Clear;
  traverse('',Owner as TForm,true);
end;

procedure TPersistor.ToForm;
begin
  fVisited.Clear;
  traverse('',Owner as TForm,false);
end;

// Read all the names and values of a registry key into
// name=Value pairs separated by crlf.
function RegKeyToStr(Path:string):string;
var reg:TRegistry; list:TStrings; i:integer;
const CRLF=#13+#10;
begin
  list:=TStringList.Create;
  reg:=TRegistry.Create;
  reg.RootKey:=HKEY_LOCAL_MACHINE;
  reg.OpenKey(path, True);
  reg.GetValueNames(list);
  for i:=0 to list.Count-1 do
    result:=result+list[i]+'='+reg.ReadString(list[i])+CRLF;
  reg.CloseKey;
  reg.Free;
  list.free;
end;

// Write a list or name=value pairs separated by crlf
// to a registry key.

procedure deleteValues(reg:TRegistry);
var list:TStrings; i:integer;
begin
  list:=TStringList.Create;
  reg.GetValueNames(list);
  for i:=0 to list.Count-1 do
    reg.DeleteValue(list[i]);
  list.free;
end;

procedure StrLstToRegKey(path:string;list:TStrings);
var reg:TRegistry;  i:integer;
begin
  reg:=TRegistry.Create;
  reg.RootKey:=HKEY_LOCAL_MACHINE;
  reg.OpenKey(path, True);
  deleteValues(reg);
  for i:=0 to list.Count-1 do
    reg.WriteString(list.Names[i],list.Values[list.Names[i]]);
  reg.CloseKey;
  reg.Free;
end;

procedure TPersistor.FromStorage;
begin
  case fStorage of
  jsInternal: fPropValues.Text:=unquote(fVault.Values[SubstFilename]);
  jsRegistry: fPropValues.Text:=RegKeyToStr(SubstFilename);
  jsFile:
    try
      fPropValues.LoadFromFile(SubstFileName);
    except
    end;
  end;
end;

procedure CreatePath(s:string);
var dir,olddir:string;
begin
  olddir:=getcurrentdir;
  while frontstring(s,dir,['\'],[]) do
  begin
    CreateDir(dir);
    chdir(dir);
  end;
  chdir(olddir);
end;

procedure TPersistor.ToStorage;
begin
  case fStorage of
  jsInternal: fVault.Values[SubstFilename]:=quote(fPropValues.Text);
  jsRegistry:strLstToRegKey(SubstFilename,fPropValues);
  jsFile:
  begin
    CreatePath(ExtractFileDir(SubstFileName));
    fPropValues.SaveToFile(SubstFileName);
  end;
  end;
end;

procedure Register;
begin
  RegisterComponents('Modularize', [TPersistor]);
  RegisterComponentEditor(TPersistor, TPersistorEditor);
end;


end.
