unit frmuPersistor;
{
  Unit type (Component, Formless unit, Ancestor form, Application form,...)
  ------------------------------------------------------------------------------
  Form for Setting Component Properties

  Purpose
  ------------------------------------------------------------------------------
  A Form to set TPersistor properties

  History
  ------------------------------------------------------------------------------
  1998-05-00 THk Created


  Copyright
  ------------------------------------------------------------------------------
  (C) 1998 - Thor Kristmundsson

}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, ExtCtrls,Persistor;

type
  TfrmPersistor = class(TForm)
    Panel1: TPanel;
    memPropValues: TMemo;
    memWhichProperties: TMemo;
    Splitter1: TSplitter;
    Persistor1: TPersistor;
    Panel3: TPanel;
    StaticText1: TStaticText;
    Image1: TImage;
    Image2: TImage;
    Image3: TImage;
    btnFromForm: TButton;
    btnApply: TButton;
    btnLoad: TButton;
    btnSave: TButton;
    StaticText2: TStaticText;
    btnCancel: TButton;
    btnOk: TButton;
    cmbUse: TComboBox;
    Label1: TLabel;
    procedure btnFromFormClick(Sender: TObject);
    procedure btnApplyClick(Sender: TObject);
    procedure btnLoadClick(Sender: TObject);
    procedure btnSaveClick(Sender: TObject);
    procedure btnOkClick(Sender: TObject);
    procedure FormCloseQuery(Sender: TObject; var CanClose: Boolean);
    procedure FormCreate(Sender: TObject);
    procedure cmbUseChange(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
    Persistor:TPersistor;
  end;

var
  frmPersistor: TfrmPersistor;

implementation
uses registry,utlStrings;
{$R *.DFM}

var Prefab:TStringList;
const rootkey:integer=integer(HKEY_LOCAL_MACHINE);

// RegexWrite writes a regsitry value with the datatype of the value. The value
// is a variant which means data type may vary. If value is of datatype other
// than integer string or boolean, such as datetime, you must use RegexRead2 to
// read it back.
function regexWrite(RootKey:integer; path, name:string; value:variant):boolean;
var reg:TRegistry;
begin
  result:=true;
  reg:=TRegistry.Create;
  reg.RootKey:=RootKey;
  reg.OpenKey(path, True);
  case varType(value) of
    varSmallint,
    varInteger,
    varByte:      reg.WriteInteger(name, value);
    varSingle,
    varDouble:    reg.WriteFloat(name, value);
    varCurrency:  reg.WriteCurrency(name, value);
    varDate:      reg.WriteDateTime(name, value);
    varBoolean:   reg.WriteBool(name, value);
    varString:    reg.WriteString(name, value);
  else
    result:=false;
  end;
  reg.CloseKey;
  reg.Free;
end;

// RegexRead returns a variant with the same datatype as the registry value to
// be read. if an error occurs or the value doesn't exist the default is
// returned.
function regexRead(RootKey:integer; path, name:string; default:variant):variant;
var reg:TRegistry;
begin
  reg:=TRegistry.Create;
  reg.RootKey:=RootKey;
  reg.OpenKey(path, True);
  case reg.GetDataType(name) of
    rdBinary,
    rdUnknown:       result:=default;
    rdString,
    rdExpandString:  result:=reg.ReadString(name);
    rdInteger:       result:=reg.ReadInteger(name);
  else
    result:=default;
  end;
  reg.CloseKey;
  reg.Free;
end;

procedure TfrmPersistor.btnFromFormClick(Sender: TObject);
begin
  Persistor.WhichProperties.Text:=memWhichProperties.Text;
  Persistor.FromForm;
  memPropValues.Text:=Persistor.fPropValues.Text;
end;

procedure TfrmPersistor.btnApplyClick(Sender: TObject);
begin
  Persistor.WhichProperties.Text:=memWhichProperties.Text;
  Persistor.fPropValues.Text:=memPropValues.Text;
  Persistor.ToForm;
end;

procedure TfrmPersistor.btnLoadClick(Sender: TObject);
begin
  Persistor.FromStorage;
  memPropValues.Text:=Persistor.fPropValues.Text;
end;

procedure TfrmPersistor.btnSaveClick(Sender: TObject);
begin
  Persistor.fPropValues.Text:=memPropValues.Text;
  Persistor.ToStorage;
end;

procedure TfrmPersistor.btnOkClick(Sender: TObject);
begin
  Persistor.WhichProperties.Text:=memWhichProperties.Text;
end;

procedure TfrmPersistor.FormCloseQuery(Sender: TObject;
  var CanClose: Boolean);
begin
  RegExWrite(rootkey,Persistor1.SubstFilename,'Left',inttostr(Left));
  RegExWrite(rootkey,Persistor1.SubstFilename,'Top',inttostr(Top));
  RegExWrite(rootkey,Persistor1.SubstFilename,'Height',inttostr(Height));
  RegExWrite(rootkey,Persistor1.SubstFilename,'Width',inttostr(Width));
  with MemWhichProperties do
    RegExWrite(rootkey,Persistor1.SubstFilename,Name+'.Height',inttostr(Height));
end;

procedure TfrmPersistor.FormCreate(Sender: TObject);
var i:integer;
begin
  for i:=0 to Prefab.Count-1 do
    cmbUse.Items.Add(Prefab.Names[i]);
  Left:=StrToInt(RegExRead(rootkey,Persistor1.SubstFilename,'Left',inttostr(Left)));
  Top:=StrToInt(RegExRead(rootkey,Persistor1.SubstFilename,'Top',inttostr(Top)));
  Height:=StrToInt(RegExRead(rootkey,Persistor1.SubstFilename,'Height',inttostr(Height)));
  Width:=StrToInt(RegExRead(rootkey,Persistor1.SubstFilename,'Width',inttostr(Width)));
  with MemWhichProperties do
    Height:=StrToInt(RegExRead(rootkey,Persistor1.SubstFilename,Name+'.Height',inttostr(Height)));
end;

procedure TfrmPersistor.cmbUseChange(Sender: TObject);
begin
  MemWhichProperties.Lines.Text:=
    strrpl(Prefab.Values[cmbUse.Items[cmbUse.ItemIndex]],'|',#13+#10)
end;

initialization
  Prefab:=TStringList.Create;
  Prefab.Add('Configuration=TMenuItem=Visible,Enabled|TWinControl=Visible,Enabled');
  Prefab.Add('Globalization=TComponent=Caption,Hint');
  Prefab.Add('Settings=TObject=Text|Memo=Lines');
  Prefab.Add('Dimensions=TComponent=Left,Top,Width,Height|TForm=WindowState');

finalization
  Prefab.Clear;
  Prefab.free;

end.
